# $Id: si.tcl 756 2006-10-10 19:29:22Z sergei $

# File transfer via Stream Initiation (XEP-0096)

###############################################################################

namespace eval si {
    set winid 0
    set chunk_size 1024

    variable options

    custom::defgroup {Stream Initiation} \
	[::msgcat::mc "Stream initiation options."] \
	-group FileTransfer

    custom::defvar options(enable) 1 \
	[::msgcat::mc "Enable SI transport for outgoing file transfers."] \
	-group {Stream Initiation} -type boolean
}

set ::NS(file-transfer) http://jabber.org/protocol/si/profile/file-transfer

###############################################################################
#
# Draw a send file dialog
#

proc si::send_file_dialog {jid args} {
    variable winid

    foreach {opt val} $args {
	switch -- $opt {
	    -connection { set connid $val }
	}
    }
    if {![info exists connid]} {
	error "[namespace current]::send_file_dialog: -connection option\
	       is mandatory"
    }

    set token [namespace current]::[incr winid]
    upvar #0 $token state

    set w .sfd$winid
    set state(w) $w
    set state(jid) $jid
    set state(connid) $connid

    Dialog $w -title [format [::msgcat::mc "Send file to %s"] $jid] \
	-separator 1 -anchor e -modal none \
	-default 0 -cancel 1

    set f [$w getframe]
    set state(f) $f

    label $f.lfile -text [::msgcat::mc "File to send:"]
    entry $f.file -textvariable ${token}(filename)
    button $f.browsefile -text [::msgcat::mc "Browse..."] \
	-command [list [namespace current]::set_send_file_name $token]

    label $f.ldesc -text [::msgcat::mc "Description:"]
    textUndoable $f.desc -width 50 -height 5 -wrap word

    ProgressBar $f.pb -variable ${token}(progress)
    set state(pb) $f.pb
    set state(progress) 0

    grid $f.lfile      -row 0 -column 0 -sticky e
    grid $f.file       -row 0 -column 1 -sticky ew
    grid $f.browsefile -row 0 -column 2 -sticky ew
    
    grid $f.ldesc -row 1 -column 0 -sticky en
    grid $f.desc  -row 1 -column 1 -sticky ewns -columnspan 2 -pady 1m

    grid $f.pb -row 2 -column 0 -sticky ew -columnspan 3 -pady 2m

    grid columnconfigure $f 1 -weight 1
    grid rowconfigure $f 1 -weight 1

    $w add -text [::msgcat::mc "Send"] \
	   -command [list [namespace current]::send_file_negotiate $token]
    $w add -text [::msgcat::mc "Cancel"] -command [list destroy $w]

    bind $w <Destroy> [list [namespace current]::send_file_close $token $w %W]

    $w draw $f.file
}

proc si::set_send_file_name {token} {
    variable $token
    upvar 0 $token state

    set file [tk_getOpenFile]
    if {$file != ""} {
	set state(filename) $file
    }
}

###############################################################################

proc si::send_file_negotiate {token} {
    upvar #0 $token state
    variable chunk_size

    $state(w) itemconfigure 0 -state disabled

    set desc [$state(f).desc get 0.0 "end -1c"]

    if {[catch {open $state(filename)} fd]} {
	ft::report_cannot_open_file $state(w) $state(filename) \
				    [ft::get_POSIX_error_desc]
	$state(w) itemconfigure 0 -state normal
	return
    }

    debugmsg filetransfer "SENDFILE: $state(filename); $desc"

    set state(fd) $fd
    fconfigure $fd -translation binary

    set name [file tail $state(filename)]
    set size [file size $state(filename)]
    set state(size) $size

    if {$size == 0} {
	$state(pb) configure -maximum 1
	set state(progress) -1
    } else {
	$state(pb) configure -maximum $size
    }

    set state(stream) [si::newout $state(connid) $state(jid)]

    set profile [jlib::wrapper:createtag file \
		     -vars [list xmlns $::NS(file-transfer) \
				 name $name \
				 size $size] \
		     -subtags [list [jlib::wrapper:createtag desc \
					 -chdata $desc]]]

    si::connect $state(stream) $chunk_size application/octet-stream \
		$::NS(file-transfer) $profile \
		[list [namespace current]::send_file $token]
}

###############################################################################

proc si::send_file {token res} {
    upvar #0 $token state

    if {![lindex $res 0]} {
	MessageDlg .auth_err -aspect 50000 -icon error \
	    -message [format [::msgcat::mc "Request failed: %s"] \
			  [lindex $res 1]] -type user \
	    -buttons ok -default 0 -cancel 0
	close $state(fd)
	$state(w) itemconfigure 0 -state normal
	return
    }

    set_status [::msgcat::mc "Transferring..."]

    after idle [list [namespace current]::send_chunk $token]
}

proc si::send_chunk {token} {
    upvar #0 $token state
    variable chunk_size

    set chunk [read $state(fd) $chunk_size]
    if {$chunk != ""} {
	si::send_data $state(stream) $chunk \
		      [list [namespace current]::send_chunk_response $token]
    } else {
	destroy $state(w)
    }
}

proc si::send_chunk_response {token res} {
    upvar #0 $token state

    if {![info exists state(stream)]} return

    if {![lindex $res 0]} {
	MessageDlg .auth_err -aspect 50000 -icon error \
	    -message [format [::msgcat::mc "Transfer failed: %s"] \
			  [lindex $res 1]] -type user \
	    -buttons ok -default 0 -cancel 0
	if {$state(size) > 0} {
	    set state(progress) 0
	}
	si::close $state(stream)
	si::freeout $state(stream)
	close $state(fd)
	$state(w) itemconfigure 0 -state normal
	return
    }

    if {$state(size) > 0} {
	set state(progress) [tell $state(fd)]
    }
    after idle [list [namespace current]::send_chunk $token]
}

###############################################################################

proc si::send_file_close {token w1 w2} {
    upvar #0 $token state

    if {$w1 != $w2} return

    catch { si::close $state(stream) }
    catch { si::freeout $state(stream) }
    catch { close $state(fd) }
    catch { unset $token }
}

###############################################################################
###############################################################################

proc si::recv_file_dialog {connid from id name size date hash desc} {
    variable winid

    set token [namespace current]::[incr winid]
    upvar #0 $token state

    set w .rfd$winid
    set state(w) $w

    set state(connid) $connid
    set state(jid) $from
    set state(id) $id

    Dialog $w -title [format [::msgcat::mc "Receive file from %s"] $from] \
	-separator 1 -anchor e \
	-modal none -default 0 -cancel 1


    set f [$w getframe]

    label $f.lname -text [::msgcat::mc "Name:"]
    label $f.name -text $name

    label $f.lsize -text [::msgcat::mc "Size:"]
    label $f.size -text $size

    label $f.ldesc -text [::msgcat::mc "Description:"]
    message $f.desc -width 10c -text $desc

    set dir $ft::options(download_dir)
    label $f.lsaveas -text [::msgcat::mc "Save as:"]
    entry $f.saveas -textvariable ${token}(filename)

    set state(dir) $dir
    set state(name) $name
    set state(filename) [file join $dir $name]
    button $f.browsefile -text [::msgcat::mc "Browse..."] \
	-command [list [namespace current]::set_receive_file_name $token]

    set state(progress) 0

    # Working around a bug in ProgressBar:
    # crash when setting PB variable while -maximum is 0:
    if {$size > 0} {
	ProgressBar $f.pb -variable ${token}(progress)
	$f.pb configure -maximum $size
	grid $f.pb  -row 4 -column 0 -sticky ew -columnspan 3 -pady 2m
    }

    grid $f.lname   -row 0 -column 0 -sticky e
    grid $f.name    -row 0 -column 1 -sticky w
    
    grid $f.lsize   -row 1 -column 0 -sticky e
    grid $f.size    -row 1 -column 1 -sticky w
    
    grid $f.ldesc   -row 2 -column 0 -sticky en
    grid $f.desc    -row 2 -column 1 -sticky ewns -columnspan 2 -pady 1m

    grid $f.lsaveas -row 3 -column 0 -sticky e
    grid $f.saveas  -row 3 -column 1 -sticky ew
    grid $f.browsefile  -row 3 -column 2 -sticky ew

    grid columnconfigure $f 1 -weight 1 -minsize 8c
    grid rowconfigure $f 2 -weight 1
    
    $w add -text [::msgcat::mc "Receive"] -command \
	[list [namespace current]::recv_file_start $token]
    $w add -text [::msgcat::mc "Cancel"] -command \
	[list [namespace current]::recv_file_cancel $token]
    
    $w draw

    # Can't avoid vwait, because this procedure must return result or error
    vwait ${token}(result)

    set result $state(result)
    if {[lindex $result 0] == "error"} {
	catch { unset $token }
    }

    return $result
}

proc si::set_receive_file_name {token} {
    upvar #0 $token state

    set file [tk_getSaveFile -initialdir $state(dir) \
			     -initialfile $state(name)]
    if {$file != ""} {
	set state(filename) $file
    }
}

###############################################################################

proc si::recv_file_cancel {token} {
    upvar #0 $token state

    set state(result) [list error cancel not-allowed]
    destroy $state(w)
}

###############################################################################

proc si::recv_file_start {token} {
    upvar #0 $token state

    if {[catch {si::newin $state(connid) $state(jid) $state(id)} stream]} {
	set state(result) [list error modify bad-request \
				-text "Stream ID is in use"]
	destroy $state(w)
	return
    }

    set state(stream) $stream

    if {[catch {open $state(filename) w} fd]} {
	ft::report_cannot_open_file $state(w) $state(filename) \
				    [ft::get_POSIX_error_desc]
	return
    }

    fconfigure $fd -translation binary

    set state(fd) $fd
    set w $state(w)

    $state(w) itemconfigure 0 -state disabled
    bind $state(w) <Destroy> [list [namespace current]::recv_file_close \
				   $token $state(w) %W]

    si::set_readable_handler \
	$stream [list [namespace current]::recv_file_chunk $token]
    si::set_closed_handler \
	$stream [list [namespace current]::closed $token]

    set state(result) {}
}

###############################################################################

proc si::recv_file_chunk {token stream} {
    upvar #0 $token state

    if {[info exists state(stream)] && $state(stream) == $stream} {
	set fd $state(fd)
	set filename $state(filename)
	set data [si::read_data $stream]

	debugmsg filetransfer "RECV into $filename data $data"

	puts -nonewline $fd $data
	set state(progress) [tell $fd]
	return 1
    } else {
	return 0
    }
}

###############################################################################

proc si::closed {token stream} {
    upvar #0 $token state

    if {[info exists state(stream)] && $state(stream) == $stream} {
	debugmsg filetransfer CLOSE
    	destroy $state(w)
    }
}

###############################################################################

proc si::recv_file_close {token w1 w2} {
    upvar #0 $token state

    if {$w1 != $w2} return

    catch {close $state(fd)}
    catch {si::freein $state(stream)}
    catch {unset $token}
}

###############################################################################

proc si::si_handler {connid from id mimetype child} {
    debugmsg filetransfer "SI set: [list $from $child]"

    jlib::wrapper:splitxml $child tag vars isempty chdata children

    if {$tag == "file"} {
	set desc ""
	foreach item $children {
	    jlib::wrapper:splitxml $item tag1 vars1 isempty1 chdata1 children1
	    switch -- $tag1 {
		desc {set desc $chdata1}
	    }
	}

	recv_file_dialog \
	    $connid \
	    $from \
	    $id \
	    [jlib::wrapper:getattr $vars name] \
	    [jlib::wrapper:getattr $vars size] \
	    [jlib::wrapper:getattr $vars date] \
	    [jlib::wrapper:getattr $vars hash] \
	    $desc
    } else {
	return [list error modify bad-request]
    }
}

si::register_profile $::NS(file-transfer) [namespace current]::si::si_handler

###############################################################################

proc si::add_menu_item {m cascad jid args} {
    variable options

    if {!$options(enable)} return

    if {$cascad} {
	set label [::msgcat::mc "via SI..."]
    } else {
	set label [::msgcat::mc "Send file via SI..."] \
    }
    $m add command -label $label \
        -command [list eval \
		      [list [namespace current]::send_file_dialog $jid] $args]

    hook::unset_flag create_filetransfer_menu_hook transport
}

hook::add create_filetransfer_menu_hook \
    [namespace current]::si::add_menu_item 10

###############################################################################

